from .royaleapi import Client as RoyaleAPI
from .official_api import Client as OfficialAPI
from .errors import *

############
# METADATA #
############

__version__ = 'v4.0.1'
__title__ = 'clashroyale'
__license__ = 'MIT'
__github__ = 'https://github.com/cgrok/clashroyale'
